from tabulate import tabulate
from .utils import VERBOSE

class SemanticAnnotation():
    """ Semantic annotation for sample types, property types or property type assignments.
    """

    def __init__(self, openbis_obj, isNew=True, **kwargs):
        self._openbis = openbis_obj
        self._isNew = isNew;
        
        self.permId = kwargs.get('permId')
        self.entityType = kwargs.get('entityType')
        self.propertyType = kwargs.get('propertyType')
        self.predicateOntologyId = kwargs.get('predicateOntologyId')
        self.predicateOntologyVersion = kwargs.get('predicateOntologyVersion')
        self.predicateAccessionId = kwargs.get('predicateAccessionId')
        self.descriptorOntologyId = kwargs.get('descriptorOntologyId')
        self.descriptorOntologyVersion = kwargs.get('descriptorOntologyVersion')
        self.descriptorAccessionId = kwargs.get('descriptorAccessionId')
        self.creationDate = kwargs.get('creationDate')

    def __dir__(self):
        return [
            'permId', 'entityType', 'propertyType', 
            'predicateOntologyId', 'predicateOntologyVersion', 
            'predicateAccessionId', 'descriptorOntologyId',
            'descriptorOntologyVersion', 'descriptorAccessionId', 
            'creationDate', 
            'save()', 'delete()' 
        ]

    def save(self):
        if self._isNew:
            self._create()
        else:
            self._update()
            
    def _create(self):
        
        creation = {
            "@type": "as.dto.semanticannotation.create.SemanticAnnotationCreation"
        }

        if self.entityType is not None and self.propertyType is not None:
            creation["propertyAssignmentId"] = {
                "@type": "as.dto.property.id.PropertyAssignmentPermId",
                "entityTypeId" : {
                    "@type": "as.dto.entitytype.id.EntityTypePermId",
                    "permId" : self.entityType,
                    "entityKind" : "SAMPLE"
                },
                "propertyTypeId" : {
                    "@type" : "as.dto.property.id.PropertyTypePermId",
                    "permId" : self.propertyType
                }
            }
        elif self.entityType is not None:
            creation["entityTypeId"] = {
                "@type": "as.dto.entitytype.id.EntityTypePermId",
                "permId" : self.entityType,
                "entityKind" : "SAMPLE"
            }
        elif self.propertyType is not None:
            creation["propertyTypeId"] = {
                "@type" : "as.dto.property.id.PropertyTypePermId",
                "permId" : self.propertyType
            }
            
        for attr in ['predicateOntologyId', 'predicateOntologyVersion', 'predicateAccessionId', 'descriptorOntologyId', 'descriptorOntologyVersion', 'descriptorAccessionId']:
            creation[attr] = getattr(self, attr)

        request = {
            "method": "createSemanticAnnotations",
            "params": [
                self._openbis.token,
                [creation]
            ]
        }
        
        response = self._openbis._post_request(self._openbis.as_v3, request)
        self._isNew = False
        self.permId = response[0]['permId']
        
        if VERBOSE: print("Semantic annotation successfully created.")
    
    def _update(self):
        
        update = {
            "@type": "as.dto.semanticannotation.update.SemanticAnnotationUpdate",
            "semanticAnnotationId" : {
                "@type" : "as.dto.semanticannotation.id.SemanticAnnotationPermId",
                "permId" : self.permId
            }
        }
        
        for attr in ['predicateOntologyId', 'predicateOntologyVersion', 'predicateAccessionId', 'descriptorOntologyId', 'descriptorOntologyVersion', 'descriptorAccessionId']:
            update[attr] = {
                "@type" : "as.dto.common.update.FieldUpdateValue",
                "isModified" : True,
                "value" : getattr(self, attr)
            }
            
        request = {
            "method": "updateSemanticAnnotations",
            "params": [
                self._openbis.token,
                [update]
            ]
        }
        
        self._openbis._post_request(self._openbis.as_v3, request)
        if VERBOSE: print("Semantic annotation successfully updated.")
    
    def delete(self, reason):
        self._openbis.delete_entity(entity='SemanticAnnotation', id=self.permId, reason=reason)
        if VERBOSE: print("Semantic annotation successfully deleted.")

    def _repr_html_(self):
        attrs = [ 'permId', 'entityType', 'propertyType', 'predicateOntologyId', 'predicateOntologyVersion', 'predicateAccessionId', 'descriptorOntologyId', 'descriptorOntologyVersion', 'descriptorAccessionId', 'creationDate',
        ]

        html = """
            <table border="1" class="dataframe">
            <thead>
                <tr style="text-align: right;">
                <th>attribute</th>
                <th>value</th>
                </tr>
            </thead>
            <tbody>
        """

        for attr in attrs:
            html += "<tr> <td>{}</td> <td>{}</td> </tr>".format(
                attr, getattr(self, attr, '')
            )

        html += """
            </tbody>
            </table>
        """
        return html
        

    def __repr__(self):
        headers = ['attribute', 'value']
        lines = []
        lines.append(['permId', self.permId])
        lines.append(['entityType', self.entityType])
        lines.append(['propertyType', self.propertyType])
        lines.append(['predicateOntologyId', self.predicateOntologyId])
        lines.append(['predicateOntologyVersion', self.predicateOntologyVersion])
        lines.append(['predicateAccessionId', self.predicateAccessionId])
        lines.append(['descriptorOntologyId', self.descriptorOntologyId])
        lines.append(['descriptorOntologyVersion', self.descriptorOntologyVersion])
        lines.append(['descriptorAccessionId', self.descriptorAccessionId])
        lines.append(['creationDate', self.creationDate])
        return tabulate(lines, headers=headers)



from .openbis_object import OpenBisObject 
from .utils import VERBOSE
from .attribute import AttrHolder
import json

class Tag(
    OpenBisObject,
    entity='tag',
    single_item_method_name='get_tag'
):
    """ 
    """

    def __dir__(self):
        return [
            'get_samples()',
            'get_experiments()',
            'get_materials()',
            'get_owner()',
        ] + super().__dir__()

    def get_owner(self):
        return self.openbis.get_person(self.owner)

    def get_samples(self):
        return self.openbis.get_samples(tags=[self.code])
        #raise ValueError('not yet implemented')

    def get_experiments(self):
        return self.openbis.get_experiments(tags=[self.code])

    def get_materials(self):
        raise ValueError('not yet implemented')




from .attribute import AttrHolder
from .openbis_object import OpenBisObject
from .utils import VERBOSE, extract_permid, extract_nested_permid,format_timestamp
from .things import Things
from pandas import DataFrame

class Group(
    OpenBisObject,
    entity='authorizationGroup',
    single_item_method_name='get_group'

):
    """ Managing openBIS authorization groups
    """

    def __dir__(self):
        return [
            'code','description','users','roleAssignments',
            'get_users()', 'set_users()', 'add_users()', 'del_users()',
            'get_roles()', 'assign_role()', 'revoke_role(techId)'
        ]

    def get_persons(self):
        """ Returns a Things object wich contains all Persons (Users)
        that belong to this group.
        """

        def create_data_frame(attrs, props, response):
            columns = ['permId', 'userId', 'firstName', 'lastName', 'email', 'space', 'registrationDate', 'active']
            persons = DataFrame(response)
            if len(persons) == 0:
                persons = DataFrame(columns=columns)
            persons['permId'] = persons['permId'].map(extract_permid)
            persons['registrationDate'] = persons['registrationDate'].map(format_timestamp)
            persons['space'] = persons['space'].map(extract_nested_permid)

            return persons[columns]

        p = Things(
            self.openbis, entity='person',
            identifier_name='permId',
            response=self._users,
            df_initializer=create_data_frame
        )
        return p

    get_users = get_persons  # Alias
    get_members = get_persons  # Alias


    def get_roles(self, **search_args):
        """ Get all roles that are assigned to this group.
        Provide additional search arguments to refine your search.

        Usage::
            group.get_roles()
            group.get_roles(space='TEST_SPACE')
        """
        return self.openbis.get_role_assignments(group=self, **search_args)

    def assign_role(self, role, **kwargs):
        """ Assign a role to this group. If no additional attribute is provided,
        roleLevel will default to INSTANCE. If a space attribute is provided,
        the roleLevel will be SPACE. If a project attribute is provided,
        roleLevel will be PROJECT.

        Usage::
            group.assign_role(role='ADMIN')
            group.assign_role(role='ADMIN', space='TEST_SPACE')

        """

        try:
            self.openbis.assign_role(role=role, group=self, **kwargs)
            if VERBOSE:
                print(
                    "Role {} successfully assigned to group {}".format(role, self.code)
                )
        except ValueError as e:
            if 'exists' in str(e):
                if VERBOSE:
                    print(
                        "Role {} already assigned to group {}".format(role, self.code)
                    )
            else:
                raise ValueError(str(e))


    def revoke_role(self, role, space=None, project=None, reason='no reason specified'):
        """ Revoke a role from this group.
        """

        techId = None
        if isinstance(role, int):
            techId = role
        else:
            query = { "role": role }
            if space is None:
                query['space'] = ''
            else:
                query['space'] = space.upper()

            if project is None:
                query['project'] = ''
            else:
                query['project'] = project.upper()

            # build a query string for dataframe
            querystr = " & ".join(
                    '{} == "{}"'.format(key, value) for key, value in query.items()
                    )
            roles = self.get_roles().df
            if len(roles) == 0:
                if VERBOSE:
                    print(f"Role {role} has already been revoked from group {self.code}")
                return
            techId = roles.query(querystr)['techId'].values[0]

        # finally delete the role assignment
        ra = self.openbis.get_role_assignment(techId)
        ra.delete(reason)
        if VERBOSE:
            print(f"Role {role} successfully revoked from group {self.code}")
        return

    def save(self):
        if self.is_new:
            request = self._new_attrs()
            resp = self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE: print("Group successfully created.")
            # re-fetch group from openBIS
            new_data = self.openbis.get_group(resp[0]['permId'], only_data=True)
            self._set_data(new_data)
            return self

        else:
            request = self._up_attrs()
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE: print("Group successfully updated.")
            # re-fetch group from openBIS
            new_data = self.openbis.get_group(self.permId, only_data=True)
            self._set_data(new_data)




from .attribute import AttrHolder
from .openbis_object import OpenBisObject 
from .utils import parse_jackson, check_datatype, split_identifier, format_timestamp, is_identifier, is_permid, nvl, VERBOSE


class Space(
    OpenBisObject,
    entity='space',
    single_item_method_name='get_space'
):
    """ managing openBIS spaces
    """

    def __dir__(self):
        """all the available methods and attributes that should be displayed
        when using the autocompletion feature (TAB) in Jupyter
        """
        return [
            'get_projects()',
            'new_project()',
            'get_project()',
            'get_experiments()',
            'get_experiment()',
            'get_collections()',
            'get_collection()',
            'get_samples()', 
            'get_sample()',
            'get_objects()',
            'get_object()',
            'delete()'
        ] + super().__dir__()

    def __str__(self):
        return self.data.get('code', None)

    def get_samples(self, **kwargs):
        return self.openbis.get_samples(space=self.code, **kwargs)
    get_objects = get_samples  # Alias

    def get_sample(self, sample_code, project_code=None):
        if is_identifier(sample_code) or is_permid(sample_code):
            return self.openbis.get_sample(sample_code)
        else:
            if project_code is None:
                return self.openbis.get_sample('/{}/{}'.format(self.code,sample_code) )
            else:
                return self.openbis.get_sample('/{}/{}/{}'.format(self.code, project_code, sample_code) )
    get_object = get_sample  # Alias

    def get_project(self, project_code):
        if is_identifier(project_code):
            return self.openbis.get_project(project_code)
        else:
            projects = self.openbis.get_projects(code=project_code, space=self.code)
            if len(projects) == 0:
                raise ValueError(f"No project named {project_code} in space {self.code}")

            return projects[0]

    def get_projects(self, **kwargs):
        return self.openbis.get_projects(space=self.code, **kwargs)

    def new_project(self, code, description=None, **kwargs):
        return self.openbis.new_project(self.code, code, description, **kwargs)

    def get_experiments(self, **kwargs):
        return self.openbis.get_experiments(space=self.code, **kwargs)
    get_collections = get_experiments

    def get_experiment(self, experiment_code):
        if is_identifier(experiment_code):
            return self.openbis.get_experiment(experiment_code)
        else:
            experiments = self.openbis.get_experiments(code=experiment_code, space=self.code)
            if len(experiments) == 0:
                raise ValueError(f"No project named {experiment_code} in space {self.code}")

            return experiments[0]
    get_collection = get_experiment

    def new_sample(self, **kwargs):
        return self.openbis.new_sample(space=self, **kwargs)




from .property import PropertyHolder
from .attribute import AttrHolder
from .openbis_object import OpenBisObject 
from .definitions import openbis_definitions
from .utils import VERBOSE

class Experiment(
    OpenBisObject,
    entity='experiment',
    single_item_method_name='get_experiment'
):
    """Handling experiment (collection) entities in openBIS 
    """

    def _set_data(self, data):
        # assign the attribute data to self.a by calling it
        # (invoking the AttrHolder.__call__ function)
        self.a(data)
        self.__dict__['data'] = data

        # put the properties in the self.p namespace (without checking them)
        for key, value in data['properties'].items():
            self.p.__dict__[key.lower()] = value

    def __str__(self):
        return self.data['code']

    def __dir__(self):
        # the list of possible methods/attributes displayed
        # when invoking TAB-completition
        return [
            'get_datasets()', 'get_samples()',
            'set_tags()', 'add_tags()', 'del_tags()',
            'add_attachment()', 'get_attachments()', 'download_attachments()',
            'save()',
            'attrs',
            'props',
        ] + super().__dir__()

    @property
    def props(self):
        return self.__dict__['p']

    @property
    def type(self):
        return self.__dict__['type']

    @type.setter
    def type(self, type_name):
        experiment_type = self.openbis.get_experiment_type(type_name)
        self.p.__dict__['_type'] = experiment_type
        self.a.__dict__['_type'] = experiment_type

    def __getattr__(self, name):
        return getattr(self.__dict__['a'], name)

    def __setattr__(self, name, value):
        if name in ['set_properties', 'add_tags()', 'del_tags()', 'set_tags()']:
            raise ValueError("These are methods which should not be overwritten")
        elif name in ['p', 'props']:
            if isinstance(value, dict):
                for p in value:
                    setattr(self.__dict__['p'], p, value[p])
            else:
                raise ValueError("please provide a dictionary for setting properties")
        else:
            setattr(self.__dict__['a'], name, value)

    def _repr_html_(self):
        html = self.a._repr_html_()
        return html

    def set_properties(self, properties):
        for prop in properties.keys():
            setattr(self.p, prop, properties[prop])

    set_props = set_properties


    def get_datasets(self, **kwargs):
        if self.permId is None:
            return None
        return self.openbis.get_datasets(experiment=self.permId, **kwargs)

    def get_projects(self, **kwargs):
        if self.permId is None:
            return None
        return self.openbis.get_project(experiment=self.permId, **kwargs)

    def get_samples(self, **kwargs):
        if self.permId is None:
            return None
        return self.openbis.get_samples(experiment=self.permId, **kwargs)

    get_objects = get_samples # Alias

    def add_samples(self, *samples):

        for sample in samples:
            if isinstance(sample, str):
                obj = self.openbis.get_sample(sample)
            else:
                # we assume we got a sample object
                obj = sample

            # a sample can only belong to exactly one experiment
            if obj.experiment is not None:
                raise ValueError(
                    "sample {} already belongs to experiment {}".format(
                        obj.code, obj.experiment
                    )
                )
            else:
                if self.is_new:
                    raise ValueError("You need to save this experiment first before you can assign any samples to it")
                else:
                    # update the sample directly
                    obj.experiment = self.identifier
                    obj.save()
                    self.a.__dict__['_samples'].append(obj._identifier)

    add_objects = add_samples # Alias

    def del_samples(self, samples):
        # TODO: implement this method using transactions

        raise ValueError("not implemented yet")
        if not isinstance(samples, list):
            samples = [samples]

        objects = [] 
        for sample in samples:
            if isinstance(sample, str):
                obj = self.openbis.get_sample(sample)
                objects.append(obj)
            else:
                # we assume we got an object
                objects.append(obj)
        self.samples = objects

    del_objects = del_samples # Alias




import base64
import os

class Attachment():
    def __init__(self, filename, title=None, description=None):
        if not os.path.exists(filename):
            raise ValueError("File not found: {}".format(filename))
        self.fileName = filename
        self.title = title
        self.description = description

    def get_data_short(self):
        return {
            "fileName": self.fileName,
            "title": self.title,
            "description": self.description,
        }

    def get_data(self):
        with open(self.fileName, 'rb') as att:
            content = att.read()
            contentb64 = base64.b64encode(content).decode()
        return {
            "fileName": self.fileName,
            "title": self.title,
            "description": self.description,
            "content": contentb64,
            "@type": "as.dto.attachment.create.AttachmentCreation",
        }



from datetime import datetime
from pathlib import Path
import re

# display messages when in a interactive context (IPython or Jupyter)
try:
    get_ipython()
except Exception:
    VERBOSE = False
else:
    VERBOSE = True

def parse_jackson(input_json):
    """openBIS uses a library called «jackson» to automatically generate the JSON RPC output.
       Objects that are found the first time are added an attribute «@id».
       Any further findings only carry this reference id.
       This function is used to dereference the output.
    """
    interesting=['tags', 'registrator', 'modifier', 'owner', 'type',
        'parents', 'children', 'containers', # 'container',
        'properties', 'experiment', 'sample',
        'project', 'space', 'propertyType', 'entityType', 'propertyType', 'propertyAssignment',
        'externalDms', 'roleAssignments', 'user', 'users', 'authorizationGroup', 'vocabulary',
        'validationPlugin', 'dataSetPermId', 'dataStore'
    ]
    found = {}
    def build_cache(graph):
        if isinstance(graph, list):
            for item in graph:
                build_cache(item)
        elif isinstance(graph, dict) and len(graph) > 0:
            for key, value in graph.items():
                if key in interesting:
                    if isinstance(value, dict):
                        if '@id' in value:
                            found[value['@id']] = value
                        build_cache(value)
                    elif isinstance(value, list):
                        for item in value:
                            if isinstance(item, dict):
                                if '@id' in item:
                                    found[item['@id']] = item
                                build_cache(item)
                elif isinstance(value, dict):
                    build_cache(value)
                elif isinstance(value, list):
                    build_cache(value)

    def deref_graph(graph):
        if isinstance(graph, list):
            for i, list_item in enumerate(graph):
                if isinstance(list_item, int):
                    # try: # TODO: use "if list_item in found" with found.get()
                    #     graph[i] = found[list_item]
                    # except KeyError:
                    #     pass
                    if list_item in found:
                        graph[i] = found.get(list_item)
                else:
                    deref_graph(list_item)
        elif isinstance(graph, dict) and len(graph) > 0:
            for key, value in graph.items():
                if key in interesting:
                    if isinstance(value, dict):
                        deref_graph(value)
                    elif isinstance(value, int):
                        graph[key] = found.get(value)
                    elif isinstance(value, list):
                        for i, list_item in enumerate(value):
                            if isinstance(list_item, int):
                                if list_item in found:
                                    value[i] = found[list_item]
                                else:
                                    value[i] = list_item
                elif isinstance(value, dict):
                    deref_graph(value)
                elif isinstance(value, list):
                    deref_graph(value)

    build_cache(input_json)
    deref_graph(found)
    deref_graph(input_json)


def check_datatype(type_name, value):
    if type_name == 'INTEGER':
        return isinstance(value, int)
    if type_name == 'BOOLEAN':
        return isinstance(value, bool)
    if type_name == 'VARCHAR':
        return isinstance(value, str)
    return True


def split_identifier(ident):
    bla = []
    bla=ident.upper().split("/")
    results = {}
    try:
        if bla[0] == '':
            bla.pop(0)
        if bla[-1] == '':
            bla.pop(-1)
        results["space"] = bla.pop(0)
        results["code"]  = bla.pop(-1)
        results["experiment"] = bla.pop(0)
    except Exception:
        pass

    return results


def format_timestamp(ts):
    if ts is None:
        return ''
    if ts != ts:  # test for NaN
        return ''
    return datetime.fromtimestamp(round(ts/1000)).strftime('%Y-%m-%d %H:%M:%S')


def is_identifier(ident):
    # assume we got a sample identifier e.g. /TEST/TEST-SAMPLE
    match = re.search('/', ident)
    if match:
        return True
    else:
        return False


def is_permid(ident):
    match = re.match('^\d+\-\d+$', ident)
    if match:
        return True
    else:
        return False


def nvl(val, string=''):
    if val is None:
        return string
    return val

def extract_permid(permid):
    if not isinstance(permid, dict):
        return str(permid)
    return permid['permId']

def extract_code(obj):
    if not isinstance(obj, dict):
        return '' if obj is None else str(obj)
    return '' if obj['code'] is None else obj['code']

def extract_downloadUrl(obj):
    if not isinstance(obj, dict):
        return '' if obj is None else str(obj)
    return '' if obj['downloadUrl'] is None else obj['downloadUrl']

def extract_name(obj):
    if not isinstance(obj, dict):
        return '' if obj is None else str(obj)
    return '' if obj['name'] is None else obj['name']

def extract_deletion(obj):
    del_objs = []
    for deleted_object in obj['deletedObjects']:
        del_objs.append({
            "reason": obj['reason'],
            "permId": deleted_object["id"]["permId"],
            "type": deleted_object["id"]["@type"]
        })
    return del_objs


def extract_attr(attr):
    def attr_func(obj):
        if isinstance(obj, dict):
            return obj.get(attr, '')
        else:
            return str(obj)
    return attr_func


def extract_identifier(ident):
    if not isinstance(ident, dict):
        return str(ident)
    return ident['identifier']

def extract_identifiers(items):
    if not items:
        return []
    try:
        return [
            data['identifier']['identifier']
            if 'identifier' in data
            else data['permId']['permId']
            for data in items
        ]
    except TypeError:
        return []

def extract_nested_identifier(ident):
    if not isinstance(ident, dict):
        return str(ident)
    return ident['identifier']['identifier']


def extract_nested_permid(permid):
    if not isinstance(permid, dict):
        return '' if permid is None else str(permid)
    return '' if permid['permId']['permId'] is None else permid['permId']['permId']

def extract_nested_permids(items):
    if not isinstance(items, list):
        return []

    return list(item['permId']['permId'] for item in items)


def extract_property_assignments(pas):
    pa_strings = []
    for pa in pas:
        if not isinstance(pa['propertyType'], dict):
            pa_strings.append(pa['propertyType'])
        else:
            pa_strings.append(pa['propertyType']['label'])
    return pa_strings


def extract_role_assignments(ras):
    ra_strings = []
    for ra in ras:
        ra_strings.append({
            "role": ra['role'],
            "roleLevel": ra['roleLevel'],
            "space": ra['space']['code'] if ra['space'] else None
        })
    return ra_strings


def extract_person(person):
    if not isinstance(person, dict):
        if person is None:
            return ''
        else:
            return str(person)
    return person['userId']

def extract_person_details(person):
    if not isinstance(person, dict):
        return str(person)
    return "{} {} <{}>".format(
        person['firstName'],
        person['lastName'],
        person['email']
    )

def extract_id(id):
    if not isinstance(id, dict):
        return str(id)
    else:
        return id['techId']

def extract_userId(user):
    if isinstance(user, list):
        return ", ".join([
            u['userId'] for u in user
        ])
    elif isinstance(user, dict):
        return user['userId']
    else:
        return str(user)

def is_number(value):
   """detects whether a given value
   is either an integer or a floating point number
   1, 2, 3,  etc.
   1.0, 2.1, etc.
   .5, .6    etc.
   """
   number_regex = re.compile(r'^(?=.)([+-]?([0-9]*)(\.([0-9]+))?)$')
   return number_regex.search(value)




from tabulate import tabulate
from texttable import Texttable
from pybis.utils import check_datatype, split_identifier, format_timestamp, is_identifier, is_permid, nvl

class PropertyHolder():

    def __init__(self, openbis_obj, type=None):
        self.__dict__['_openbis'] = openbis_obj
        self.__dict__['_property_names'] = {}
        if type is None:
            return

        self.__dict__['_type'] = type
        if 'propertyAssignments' in type.data \
        and type.data['propertyAssignments'] is not None:
            for prop in type.data['propertyAssignments']:
                property_name = prop['propertyType']['code'].lower()
                self._property_names[property_name]=prop['propertyType']
                self._property_names[property_name]['mandatory'] = prop['mandatory']
                self._property_names[property_name]['showInEditView'] = prop['showInEditView']
                if prop['propertyType']['dataType'] == 'CONTROLLEDVOCABULARY':
                    pt = self._openbis.get_property_type(prop['propertyType']['code'])
                    # get the vocabulary of a property type.
                    # In some cases, the «code» of an assigned property is not identical to the «vocabulary» attribute
                    voc = self._openbis.get_vocabulary(pt.vocabulary)
                    terms = voc.get_terms()
                    self._property_names[property_name]['terms'] = terms

    def _all_props(self):
        props = {}
        if not getattr(self, '_type'):
            return props
        for code in self._type.codes():
            props[code] = getattr(self, code)
        return props

    def all(self):
        """Returns the properties as an array
        """
        props = {}
        for code in self._type.codes():
            props[code] = getattr(self, code)
        return props

    def all_nonempty(self):
        props = {}
        for code in self._type.codes():
            value = getattr(self, code)
            if value is not None:
                props[code] = value
        return props

    def __call__(self, *args):
        if len(args) == 0:
            return self.all()
        elif len(args) == 1:
            return getattr(self, args[0])
        elif len(args) == 2:
            return setattr(self, args[0], args[1])
        else:
            raise ValueError('called properties with more than 2 arguments')


    def get(self, *args):
        if len(args) == 0:
            return self.all()
        elif len(args) == 1 and not isinstance(args[0], list):
            return getattr(self, args[0])
        else:
            if isinstance(args[0], list):
                args = args[0]
            return {arg: getattr(self, arg, None) for arg in args}


    def set(self, *args):
        if len(args) == 2:
            setattr(self, args[0], args[1])
        elif len(args) == 1 and isinstance(args[0], dict):
            for key in args[0]:
                setattr(self, key, args[0][key])

    def __getitem__(self, key):
        """For properties that contain either a dot or a dash or any other non-valid method character,
        a user can use a key-lookup instead, e.g. sample.props['my-weird.property-name']
        """

        return getattr(self, key)

    def __getattr__(self, name):
        """ attribute syntax can be found out by
            adding an underscore at the end of the property name
        """ 
        if name == '_ipython_canary_method_should_not_exist_':
            # make Jupyter use the _repr_html_ method
            return
        if name.endswith('_'):
            name = name.rstrip('_')
            if name in self._property_names:
                property_type = self._property_names[name]
                if property_type['dataType'] == 'CONTROLLEDVOCABULARY':
                    return property_type['terms']
                    #return self._get_terms(property_type['code'])
                else:
                    syntax = { property_type["label"] : property_type["dataType"]}
                    if property_type["dataType"] == "TIMESTAMP":
                        syntax['syntax'] = 'YYYY-MM-DD HH:MIN:SS'
                    return syntax
            else:
                return

    def __setattr__(self, name, value):
        """This special method allows a PropertyHolder object
        to check the attributes that are assigned to that object
        """
        if name not in self._property_names:
            raise KeyError(
                "No such property: '{}'. Allowed properties are: {}".format(
                    name, ", ".join(self._property_names.keys())
                )
            )
        property_type = self._property_names[name]
        data_type = property_type['dataType']
        if data_type == 'CONTROLLEDVOCABULARY':
            terms = property_type['terms']
            value = str(value).upper()
            if value not in terms.df['code'].values:
                raise ValueError("Value for attribute {} must be one of these terms: {}".format(
                    name, ", ".join(terms.df['code'].values)
                ))
        elif data_type in ('INTEGER', 'BOOLEAN', 'VARCHAR'):
            if not check_datatype(data_type, value):
                raise ValueError("Value must be of type {}".format(data_type))
        self.__dict__[name] = value

    def __setitem__(self, key, value):
        """For properties that contain either a dot or a dash or any other non-valid method character,
        a user can use a key instead, e.g. sample.props['my-weird.property-name']
        """
        return setattr(self, key, value)

    def __dir__(self):
        return self._property_names

    def _repr_html_(self):
        def nvl(val, string=''):
            if val is None:
                return string
            elif val == 'true':
                return True
            elif val == 'false':
                return False
            return val
        html = """
            <table border="1" class="dataframe">
            <thead>
                <tr style="text-align: right;">
                <th>property</th>
                <th>value</th>
                <th>description</th>
                <th>type</th>
                <th>mandatory</th>
                </tr>
            </thead>
            <tbody>
        """

        for prop_name, prop in self._property_names.items():
            html += "<tr> <td>{}</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td> </tr>".format(
                prop_name, nvl(getattr(self, prop_name, ''),''),
                prop.get('description'),
                prop.get('dataType'),
                prop.get('mandatory'),
            )

        html += """
            </tbody>
            </table>
        """
        return html

    def __repr__(self):
        def nvl(val, string=''):
            if val is None:
                return string
            elif val == 'true':
                return True
            elif val == 'false':
                return False
            return str(val)

        headers = ['property', 'value', 'mandatory']

        lines = []
        for prop_name in self._property_names:
            lines.append([
                prop_name,
                nvl(getattr(self, prop_name, ''))
            ])
        return tabulate(lines, headers=headers)




from .attribute import AttrHolder
from .definitions import openbis_definitions, get_type_for_entity, get_method_for_entity
from .openbis_object import OpenBisObject
from .utils import VERBOSE


class Vocabulary(
    OpenBisObject,
    entity='vocabulary',
    single_item_method_name='get_vocabulary'
):

    def __init__old_(self, openbis_obj, data=None, terms=None, **kwargs):
        self.__dict__['openbis'] = openbis_obj
        self.__dict__['a'] = AttrHolder(openbis_obj, 'vocabulary')

        if data is not None:
            self._set_data(data)
            self.__dict__['terms'] = data['terms']

        if terms is None:
            self.__dict__['terms'] = []
        else:
            self.__dict__['terms'] = terms

        if self.is_new:
            allowed_attrs = openbis_definitions(self.entity)['attrs_new']
            for key in kwargs:
                if key not in allowed_attrs:
                    raise ValueError(
                        "{} is an unknown Vocabulary attribute. Allowed attributes are: {}".format(
                            key, ", ".join(allowed_attrs) 
                        ) 
                    )

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    def __dir__(self):
        return [
            'get_terms()',
            'add_term(code, label, description)',
            'save()',
        ]

    def get_terms(self):
        """ Returns the VocabularyTerms of the given Vocabulary.
        """
        return self.openbis.get_terms(vocabulary=self.code)


    def add_term(self, code, label=None, description=None):
        """ Adds a term to this Vocabulary.
        If Vocabulary is already persistent, it is added by adding a new VocabularyTerm object.
        If Vocabulary is new, the term is added to the list of terms
        """
        self.__dict__['terms'].append({
            "code": code,
            "label": label,
            "description": description
        })

    def delete(self, reason):
        """Delete this vocabulary
        """
        if not self.data:
            return

        delete_type = get_type_for_entity('vocabulary', 'delete')
        method = get_method_for_entity('vocabulary', 'delete')

        request = {
           "method": method,
           "params": [
                self.openbis.token,
                [{
                    "permId": self.code,
                    "@type": "as.dto.vocabulary.id.VocabularyPermId"
                }],
                {
                    "reason": reason,
                    **delete_type
                }
            ]
        }
        resp = self.openbis._post_request(self.openbis.as_v3, request)
        if VERBOSE: print(
            "{} {} successfully deleted.".format(
                self.entity,
                self.code
            )
        )
        

    def save(self):
        terms = self._terms or []
        for term in terms:
            term["@type"]= "as.dto.vocabulary.create.VocabularyTermCreation"

        if self.is_new:
            request = self._new_attrs('createVocabularies')
            if terms:
                request['params'][1][0]['terms'] = terms 
            resp = self.openbis._post_request(self.openbis.as_v3, request)

            if VERBOSE: print("Vocabulary successfully created.")
            data = self.openbis.get_vocabulary(resp[0]['permId'], only_data=True)
            self._set_data(data)
            return self

        else:
            request = self._up_attrs('updateVocabularies')
            request['params'][1][0]['vocabularyId'] = {
                "@type": "as.dto.vocabulary.id.VocabularyPermId",
                "permId" : self.code
            }
            request['params'][1][0]['chosenFromList'] = {
                "value": self.chosenFromList,
                "isModified": True,
                "@type": "as.dto.common.update.FieldUpdateValue"
            }
            if terms:
                request['params'][1][0]['terms'] = terms 
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE: print("Vocabulary successfully updated.")
            data = self.openbis.get_vocabulary(self.code, only_data=True)
            self._set_data(data)


class VocabularyTerm(OpenBisObject):

    def __init__(self, openbis_obj, data=None, **kwargs):
        self.__dict__['openbis'] = openbis_obj
        self.__dict__['a'] = AttrHolder(openbis_obj, 'vocabularyTerm')

        if data is not None:
            self._set_data(data)

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    @property
    def vocabularyCode(self):
        if self.is_new:
            return self.__dict__['a'].vocabularyCode
        else:
            return self.data['permId']['vocabularyCode']

    def __dir__(self):
        return [
            'code',
            'vocabularyCode',
            'label',
            'description',
            'official',
            'ordinal',
            'registrator',
            'registrationDate',
            'modifier',
            'modificationDate',
            'move_to_top()',
            'move_after_term()',
        ]

    def move_to_top(self):
        """ Moves the term on the top of the vocabularyTerm list,
        i.e. the ordinal will change
        """
        self.previousTermId = ""

    def move_after_term(self, term):
        """ Moves the term just after the term given. This will result in an ordinal change.
        """
        self.previousTermId = term

    def _up_attrs(self):
        """ AttributeTerms behave quite differently to all other openBIS entities,
        that's why we need to override this method
        """
        attrs = {}
        for attr in 'label description official'.split():
            attrs[attr] = {
                "value": getattr(self, attr),
                "isModified": True,
                "@type": "as.dto.common.update.FieldUpdateValue"
            }

        if not getattr(self, 'previousTermId') == None:
            # internal attribute previousTermId has been touched:
            # set value to either None (i.e. move it to the top) or
            # set value to PREVIOUS_TERM like this:
            value = self.previousTermId
            if value == "":
                value = None
            else:
                permId = {
                    '@type'         : 'as.dto.vocabulary.id.VocabularyTermPermId',
                    'vocabularyCode': self.vocabularyCode,
                    'code'          : value 
                }
                value = permId

            attrs['previousTermId'] = {
                "isModified": True,
                "@type": "as.dto.common.update.FieldUpdateValue",
                "value": value,
            }

        attrs["vocabularyTermId"] = self.vocabularyTermId()
        attrs["@type"] = "as.dto.vocabulary.update.VocabularyTermUpdate"
        request = {
            "method": "updateVocabularyTerms",
            "params": [
                self.openbis.token,
                [attrs]
            ]
        }
        return request


    def _new_attrs(self):
        attrs = {
            "@type": "as.dto.vocabulary.create.VocabularyTermCreation",
            "vocabularyId": self.vocabularyTermId()
        }
        for attr in 'code label description'.split():
            attrs[attr] = getattr(self, attr)

        request = {
            "method": "createVocabularyTerms",
            "params": [
                self.openbis.token,
                [attrs]
            ]
        }
        return request


    def vocabularyTermId(self):
        """ needed for updating a term.
        """
        if self.is_new:
            return {
                "permId": getattr(self, 'vocabularyCode'),
                "@type": "as.dto.vocabulary.id.VocabularyPermId"
            }
        else:
            permId = self.data['permId']
            permId.pop('@id', None)
            return permId


    def save(self):
        if self.is_new:
            request = self._new_attrs()
            resp = self.openbis._post_request(self.openbis.as_v3, request)

            if VERBOSE: print("Vocabulary Term successfully created.")
            data = self.openbis.get_term(
                code=resp[0]['code'], 
                vocabularyCode=resp[0]['vocabularyCode'],
                only_data=True
            )
            self._set_data(data)
            return self

        else:
            request = self._up_attrs()
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE: print("Vocabulary Term successfully updated.")
            data = self.openbis.get_term(
                code=self.code, 
                vocabularyCode=self.vocabularyCode,
                only_data=True
            )
            self._set_data(data)


    def delete(self, reason='no particular reason'):
        self.openbis.delete_openbis_entity(
            entity='VocabularyTerm', objectId=self.data['permId'], reason=reason
        )
        if VERBOSE: print("VocabularyTerm successfully deleted.")



from .attribute import AttrHolder
from .openbis_object import OpenBisObject
from .property import PropertyHolder
from .utils import VERBOSE


class Material(OpenBisObject):
    """ Managing openBIS materials
    """

    def __init__(self, openbis_obj, type, data=None, props=None, **kwargs):
        self.__dict__['entity'] = 'material'
        self.__dict__['openbis'] = openbis_obj
        self.__dict__['type'] = type
        self.__dict__['p'] = PropertyHolder(openbis_obj, type)
        self.__dict__['a'] = AttrHolder(openbis_obj, 'material', type)

        if data is not None:
            self._set_data(data)

        if props is not None:
            for key in props:
                setattr(self.p, key, props[key])

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])


    def __dir__(self):
        return [
            'code','description',
            'set_tags()', 'add_tags()', 'del_tags()'
        ]


    def save(self):
        for prop_name, prop in self.props._property_names.items():
            if prop['mandatory']:
                if getattr(self.props, prop_name) is None or getattr(self.props, prop_name) == "":
                    raise ValueError("Property '{}' is mandatory and must not be None".format(prop_name))

        props = self.p._all_props()

        if self.is_new:
            request = self._new_attrs()
            request["params"][1][0]["properties"] = props
            resp = self.openbis._post_request(self.openbis.as_v3, request)

            if VERBOSE: print("Material successfully created.")
            new_material_data = self.openbis.get_tag(resp[0]['permId'], only_data=True)
            self._set_data(new_material_data)
            return self

        else:
            request = self._up_attrs()
            request["params"][1][0]["properties"] = props
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE: print("Material successfully updated.")
            new_material_data = self.openbis.get_tag(self.permId, only_data=True)
            self._set_data(new_material_data)


    def delete(self, reason='no reason'):
        self.openbis.delete_entity(entity=self.entity,id=self.permId, reason=reason)
        if VERBOSE: print("Material {} successfully deleted.".format(self.permId))




from .attribute import AttrHolder
from .openbis_object import OpenBisObject
from .utils import VERBOSE


class Person(OpenBisObject):
    """managing openBIS persons"""

    def __init__(self, openbis_obj, data=None, **kwargs):
        self.__dict__["openbis"] = openbis_obj
        self.__dict__["a"] = AttrHolder(openbis_obj, "person")

        if data is not None:
            self.a(data)
            self.__dict__["data"] = data

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    def __dir__(self):
        """all the available methods and attributes that should be displayed
        when using the autocompletion feature (TAB) in Jupyter
        """
        return [
            "permId",
            "userId",
            "firstName",
            "lastName",
            "email",
            "registrator",
            "registrationDate",
            "space",
            "get_roles()",
            "assign_role(role, space)",
            "revoke_role(role)",
        ]

    def get_roles(self, **search_args):
        """Get all roles that are assigned to this person.
        Provide additional search arguments to refine your search.

        Usage::
            person.get_roles()
            person.get_roles(space='TEST_SPACE')
        """
        return self.openbis.get_role_assignments(person=self, **search_args)

    def assign_role(self, role, **kwargs):
        try:
            self.openbis.assign_role(role=role, person=self, **kwargs)
            if VERBOSE:
                print(
                    "Role {} successfully assigned to person {}".format(
                        role, self.userId
                    )
                )
        except ValueError as e:
            if "exists" in str(e):
                if VERBOSE:
                    print(
                        "Role {} already assigned to person {}".format(
                            role, self.userId
                        )
                    )
            else:
                raise ValueError(str(e))

    def revoke_role(self, role, space=None, project=None, reason="no reason specified"):
        """Revoke a role from this person."""

        techId = None
        if isinstance(role, int):
            techId = role
        else:
            query = {"role": role}
            if space is None:
                query["space"] = ""
            else:
                if isinstance(space, str):
                    query["space"] = space.upper()
                else:
                    query["space"] = space.code.upper()

            if project is None:
                query["project"] = ""
            else:
                if isinstance(project, str):
                    query["project"] = project.upper()
                else:
                    query["project"] = project.code.upper()

            # build a query string for dataframe
            querystr = " & ".join(
                '{} == "{}"'.format(key, value) for key, value in query.items()
            )
            roles = self.get_roles().df
            if len(roles) == 0:
                if VERBOSE:
                    print(
                        f"Role {role} has already been revoked from person {self.code}"
                    )
                return
            techId = roles.query(querystr)["techId"].values[0]

        # finally delete the role assignment
        ra = self.openbis.get_role_assignment(techId)
        ra.delete(reason)
        if VERBOSE:
            print("Role {} successfully revoked from person {}".format(role, self.code))
        return

    def __str__(self):
        return "{} {}".format(self.get("firstName"), self.get("lastName"))

    def delete(self, reason):
        raise ValueError("Persons cannot be deleted")

    def save(self):
        if self.is_new:
            request = self._new_attrs()
            resp = self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("Person successfully created.")
            new_person_data = self.openbis.get_person(resp[0]["permId"], only_data=True)
            self._set_data(new_person_data)
            return self

        else:
            request = self._up_attrs()
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("Person successfully updated.")
            new_person_data = self.openbis.get_person(self.permId, only_data=True)
            self._set_data(new_person_data)



from tabulate import tabulate
from pandas import DataFrame, Series
import pandas as pd

class Things():
    """An object that contains a DataFrame object about an entity  available in openBIS.
    Use .df to work with the DataFrame instead.
    Can be used in a for-loop:

    for sample in openbis.get_samples():
        ...

    You can access an element directly by providing the index number:
        openbis.get_samples()[7]

    Because the order of the elements cannot be ensured, you should choose the identifier instead:
        openbis.get_samples()['/SOME_SPACE/SAMPLE_CODE']

    Of course, if you know the identifier already, you would rather do:
        openbis.get_sample('/SOME_SPACE/SAMPLE_CODE')


    """

    def __init__(
        self, openbis_obj, entity,
        identifier_name='code', additional_identifier=None,
        start_with=None, count=None, totalCount=None,
        single_item_method=None,
        response=None,
        df_initializer=None,
        objects_initializer=None,
        attrs=None,
        props=None
    ):
        self.openbis = openbis_obj
        self.entity = entity
        self.__df = None
        self.identifier_name = identifier_name
        self.additional_identifier = additional_identifier
        self.start_with = start_with
        self.count = count
        self.totalCount=totalCount
        self.single_item_method=single_item_method
        self.__objects = None
        self.response = response
        self.__objects_initializer = objects_initializer
        self.__df_initializer = df_initializer
        self.__attrs = attrs
        self.__props = props

    def is_df_initialised(self):
        return self.__df is not None

    def is_objects_initialised(self):
        return self.__objects is not None

    @property
    def df(self):
        if self.__df is None and self.__df_initializer is not None:
            self.__df = self.__df_initializer(attrs=self.__attrs, props=self.__props, response=self.response)
        return self.__df

    @property
    def objects(self):
        if self.__objects is None and self.__objects_initializer is not None:
            self.__objects = self.__objects_initializer(response=self.response)
        return self.__objects

    def __repr__(self):
        return tabulate(self.df, headers=list(self.df))

    def __len__(self):
        return len(self.df)

    def _repr_html_(self):
        return self.df._repr_html_()

    @staticmethod
    def __create_data_frame(attrs, props, response):
        if len(response) > 0:
            return pd.concat(response)
        else:
            return DataFrame()

    def get_parents(self, **kwargs):
        if self.entity not in ['sample', 'dataset']:
            raise ValueError("{}s do not have parents".format(self.entity))

        if self.df is not None and len(self.df) > 0:
            dfs = []
            for ident in self.df[self.identifier_name]:
                # get all objects that have this object as a child == parent
                try:
                    parents = getattr(self.openbis, 'get_' + self.entity.lower() + 's')(withChildren=ident, **kwargs)
                    dfs.append(parents.df)
                except ValueError:
                    pass
            return Things(self.openbis, self.entity, self.identifier_name, response=dfs,
                          df_initializer=self.__create_data_frame)


    def get_children(self, **kwargs):
        if self.entity not in ['sample', 'dataset']:
            raise ValueError("{}s do not have children".format(self.entity))

        if self.df is not None and len(self.df) > 0:
            dfs = []
            for ident in self.df[self.identifier_name]:
                # get all objects that have this object as a child == parent
                try:
                    parents = getattr(self.openbis, 'get_' + self.entity.lower() + 's')(withParent=ident, **kwargs)
                    dfs.append(parents.df)
                except ValueError:
                    pass

                return Things(self.openbis, self.entity, self.identifier_name, response=dfs,
                              df_initializer=self.__create_data_frame)

    def get_samples(self, **kwargs):
        if self.entity not in ['space', 'project', 'experiment']:
            raise ValueError("{}s do not have samples".format(self.entity))

        if self.df is not None and len(self.df) > 0:
            dfs = []
            for ident in self.df[self.identifier_name]:
                args = {}
                args[self.entity.lower()] = ident
                try:
                    samples = self.openbis.get_samples(**args, **kwargs)
                    dfs.append(samples.df)
                except ValueError:
                    pass

            return Things(self.openbis, 'sample', 'identifier', response=dfs,
                          df_initializer=self.__create_data_frame)

    get_objects = get_samples # Alias

    def get_datasets(self, **kwargs):
        if self.entity not in ['sample', 'experiment']:
            raise ValueError("{}s do not have datasets".format(self.entity))

        if self.df is not None and len(self.df) > 0:
            dfs = []
            for ident in self.df[self.identifier_name]:
                args = {}
                args[self.entity.lower()] = ident
                try:
                    datasets = self.openbis.get_datasets(**args, **kwargs)
                    dfs.append(datasets.df)
                except ValueError:
                    pass

            return Things(self.openbis, 'dataset', 'permId', response=dfs,
                          df_initializer=self.__create_data_frame)

    def __getitem__(self, key):
        """ elegant way to fetch a certain element from the displayed list.
        If an integer value is given, we choose the row.
        If the key is a list, we return the desired columns (normal dataframe behaviour)
        If the key is a non-integer value, we treat it as a primary-key lookup
        """
        if self.df is not None and len(self.df) > 0:
            row = None
            if isinstance(key, int):
                if self.objects:
                    return self.objects[key]
                else:
                    # get thing by rowid
                    row = self.df.loc[[key]]
            elif isinstance(key, list):
                # treat it as a normal dataframe
                return self.df[key]
            else:
                # get thing by code
                row = self.df[self.df[self.identifier_name] == key.upper()]

            if row is not None:
                # invoke the openbis.get_<entity>() method
                if self.single_item_method:
                    get_item = self.single_item_method
                else:
                    get_item = getattr(self.openbis, 'get_' + self.entity)
                if self.additional_identifier is None:
                    return get_item(row[self.identifier_name].values[0])
                ## get an entry using two keys
                else:
                    return get_item(
                            row[self.identifier_name].values[0],
                            row[self.additional_identifier].values[0]
                    )

    def __iter__(self):
        if self.objects:
            for obj in self.objects:
                yield obj
        else:
            if self.single_item_method:
                get_item = self.single_item_method
            else:
                get_item = getattr(self.openbis, 'get_' + self.entity)
            for item in self.df[[self.identifier_name]][self.identifier_name].iteritems():
                yield get_item(item[1])
                #yield getattr(self.openbis, 'get_' + self.entity)(item[1])

                # return self.df[[self.identifier_name]].to_dict()[self.identifier_name]




from .attribute import AttrHolder
from .openbis_object import OpenBisObject 
from .utils import VERBOSE


class Project(
    OpenBisObject,
    entity='project',
    single_item_method_name='get_project'
):

    def _modifiable_attrs(self):
        return

    def __dir__(self):
        """all the available methods and attributes that should be displayed
        when using the autocompletion feature (TAB) in Jupyter
        """

        return [
            'add_attachment()',
            'get_attachments()', 'download_attachments()',
            'get_experiments()', 'get_samples()', 'get_datasets()',
            'save()', 'delete()'
        ] + super().__dir__()

    @property
    def props(self):
        return self.__dict__['p']

    def get_samples(self, **kwargs):
        return self.openbis.get_samples(project=self.permId, **kwargs)
    get_objects = get_samples # Alias

    def get_sample(self, sample_code):
        if is_identifier(sample_code) or is_permid(sample_code):
            return self.openbis.get_sample(sample_code)
        else:
            # we assume we just got the code
            return self.openbis.get_sample(project=self, code=sample_code)
    get_object = get_sample # Alias


    def get_experiments(self):
        return self.openbis.get_experiments(project=self.permId)
    get_collections = get_experiments  # Alias

    def get_datasets(self):
        return self.openbis.get_datasets(project=self.permId)



from tabulate import tabulate
from texttable import Texttable
from pandas import DataFrame
from .openbis_object import OpenBisObject
from .things import Things
from .utils import (
    check_datatype,
    split_identifier,
    format_timestamp,
    is_identifier,
    is_permid,
    nvl,
    extract_permid,
    extract_code,
    extract_name,
    VERBOSE,
)
from .definitions import (
    get_method_for_entity,
    get_type_for_entity,
    get_definition_for_entity,
)
from .semantic_annotation import SemanticAnnotation


class EntityType:
    """EntityTypes define a variety of an entity, eg. sample, dataSet, experiment
    This is the parent class of the SampleType, DataSetType, ExperimentType and
    MaterialType classes.
    """

    def __init__(self, openbis_obj, data=None, method=None, **kwargs):
        """This __init__ is called by OpenBisObject.__init__
        It stores the propertyAssignments data into the _propertyAssignments
        dict
        """
        self._set_entity_data(data)
        if method:
            self.__dict__["_get_method"] = method

    def _set_entity_data(self, data=None):
        pas = []
        if data is not None and "propertyAssignments" in data:
            pas = data["propertyAssignments"]
        self.__dict__["_propertyAssignments"] = pas

    def __str__(self):

        """String representation of this entity type"""
        return self.data["code"]

    def _attrs(self):
        return [
            "code",
            "description",
            "autoGeneratedCode",
            "subcodeUnique",
            "generatedCodePrefix",
            "listable",
            "showContainer",
            "showParents",
            "showParentMetadata",
            "validationPlugin",
        ]

    def __dir__(self):
        defs = get_definition_for_entity(self.entity)
        attrs = [
            "get_property_assignments()",
            "assign_property()",
            "revoke_property()",
            "move_property_to_top()",
            "move_property_after()",
            "get_validationPlugin()",
            "save()",
            "delete()",
            "get_next_sequence()",
            "get_next_code()",
        ]
        if self.is_new:
            return attrs + defs["attrs_new"]
        else:
            return attrs + list(set(defs["attrs"] + defs["attrs_up"]))

    def __getattr__(self, name):
        if name in self._attrs():
            if name in self.data:
                return self.data[name]
            else:
                return ""

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    def get_property_assignments(self):
        attrs = [
            "propertyType",
            "section",
            "ordinal",
            "mandatory",
            "initialValueForExistingEntities",
            "showInEditView",
            "showRawValueInForms",
            "registrator",
            "registrationDate",
            "plugin",
        ]

        pas = self.__dict__["_propertyAssignments"]

        def create_data_frame(attrs, props, response):
            df = DataFrame(response, columns=attrs)

            if "propertyType" in df:
                df["propertyType"] = df["propertyType"].map(extract_code)

            if "plugin" in df:
                df["plugin"] = df["plugin"].map(extract_name)

            if "registrationDate" in df:
                df["registrationDate"] = df["registrationDate"].map(format_timestamp)

            return df

        return Things(
            openbis_obj=self.openbis,
            entity="propertyType",
            single_item_method=self.openbis.get_property_type,
            identifier_name="propertyType",
            start_with=1,
            count=len(pas),
            totalCount=len(pas),
            response=pas,
            df_initializer=create_data_frame,
        )

    def assign_property(
        self,
        prop,
        plugin=None,
        section=None,
        ordinal=None,
        mandatory=False,
        initialValueForExistingEntities=None,
        showInEditView=True,
        showRawValueInForms=True,
    ):
        """The «section» groups certain properties.
        The «ordinal» is defining the rank in the list where the property appears.
        The «mandatory» defines whether a property must be filled in. If you make a
        property mandatory after you already created entities, you have to define an
        «initialValueForExistingEntities» too.
        """
        if self.is_new:
            raise ValueError("Please save {} first".format(self.entity))

        if isinstance(prop, str):
            property_type = self.openbis.get_property_type(prop.upper())
        else:
            property_type = prop

        new_assignment = {
            "section": section,
            "ordinal": ordinal,
            "mandatory": mandatory,
            "initialValueForExistingEntities": initialValueForExistingEntities,
            "showInEditView": showInEditView,
            "showRawValueInForms": showRawValueInForms,
            "propertyTypeId": {
                "@type": "as.dto.property.id.PropertyTypePermId",
                "permId": property_type.permId,
            },
            "@type": "as.dto.property.create.PropertyAssignmentCreation",
        }

        # assign plugin
        if plugin is not None:
            plugin_obj = self.openbis.get_plugin(plugin)
            new_assignment["plugin"] = plugin_obj.name

        request = self._get_request_for_pa(new_assignment, "Add")
        try:
            self.openbis._post_request(self.openbis.as_v3, request)
        except ValueError as exc:
            if "already assigned" in str(exc):
                if VERBOSE:
                    print(
                        f"Property {property_type.permId} already assigned to {self.permId}"
                    )
                return
            else:
                raise ValueError(exc)

        new_data = self._get_method(self.permId, only_data=True)
        self._set_entity_data(new_data)
        if VERBOSE:
            print(f"Property {property_type.permId} assigned to {self.permId}")

    def revoke_property(self, prop, force=False):
        if isinstance(prop, str):
            property_type = prop.upper()
        else:
            property_type = prop.permId.upper()

        items = {
            "entityTypeId": self._permId,
            "propertyTypeId": {
                "permId": property_type,
                "@type": "as.dto.property.id.PropertyTypePermId",
            },
            "@type": "as.dto.property.id.PropertyAssignmentPermId",
        }
        request = self._get_request_for_pa(items, "Remove", force)
        resp = self.openbis._post_request(self.openbis.as_v3, request)
        if not resp and VERBOSE:
            new_data = self._get_method(self.permId, only_data=True)
            self._set_entity_data(new_data)

            print(f"Property {property_type} revoked from {self.permId}")

    def _get_request_for_pa(self, items, item_action, force=False):

        method = get_method_for_entity(self.entity, "update")
        entity_type = get_type_for_entity(self.entity, "update")

        list_action_type = "as.dto.common.update.ListUpdateAction" + item_action

        request = {
            "method": method,
            "params": [
                self.openbis.token,
                [
                    {
                        "propertyAssignments": {
                            "actions": [
                                {
                                    "items": [
                                        {
                                            **items,
                                        }
                                    ],
                                    "@type": list_action_type,
                                }
                            ],
                            "@type": "as.dto.entitytype.update.PropertyAssignmentListUpdateValue",
                            "forceRemovingAssignments": force,
                        },
                        **entity_type,
                        "typeId": self._permId,
                    }
                ],
            ],
        }

        return request

    def move_property_to_top(self, property):
        raise ValueError("not implemented yet")

    def move_property_after(self, property, after_property):
        raise ValueError("not implemented yet")

    @property
    def get_validationPlugin(self):
        """Returns a validation plugin object when called.
        Returns None when no validation plugin is defined.
        """
        try:
            return self.openbis.get_plugin(self._validationPlugin["name"])
        except Exception:
            pass

    def codes(self):
        codes = []
        for pa in self.data["propertyAssignments"]:
            codes.append(pa["propertyType"]["code"].lower())
        return codes

    def get_next_sequence(self):
        request = {
            "method": "executeCustomASService",
            "params": [
                self.openbis.token,
                {
                    "permId": "as-eln-lims-api",
                    "@type": "as.dto.service.id.CustomASServiceCode",
                },
                {
                    "parameters": {
                        "method": "getNextSequenceForType",
                        "sampleTypeCode": self.code,
                    },
                    "@type": "as.dto.service.CustomASServiceExecutionOptions",
                },
            ],
        }

        try:
            resp = self.openbis._post_request(self.openbis.as_v3, request)
        except Exception:
            return None
        return resp

    def get_next_code(self):
        """Returns the next possible code for a new instance for this entity type."""
        seq = self.get_next_sequence()
        if seq:
            return self.generatedCodePrefix + str(seq)
        else:
            raise ValueError("Could not generate next code.")


class SampleType(
    OpenBisObject,
    EntityType,
    entity="sampleType",
    single_item_method_name="get_sample_type",
):
    def __init__(
        self, openbis_obj, type=None, data=None, props=None, method=None, **kwargs
    ):
        OpenBisObject.__init__(
            self, openbis_obj, type=type, data=data, props=props, **kwargs
        )
        EntityType.__init__(
            self,
            openbis_obj,
            type=type,
            data=data,
            props=props,
            method=method,
            **kwargs,
        )

    def __dir__(self):
        return (
            ["add_semantic_annotation()", "get_semantic_annotations()"]
            + EntityType.__dir__(self)
            + OpenBisObject.__dir__(self)
        )

    def add_semantic_annotation(self, **kwargs):
        semantic_annotation = SemanticAnnotation(
            openbis_obj=self.openbis, isNew=True, entityType=self.code, **kwargs
        )
        semantic_annotation.save()
        return semantic_annotation

    def get_semantic_annotations(self):
        return self.openbis.search_semantic_annotations(entityType=self.code)


class DataSetType(
    OpenBisObject,
    EntityType,
    entity="dataSetType",
    single_item_method_name="get_dataset_type",
):
    def __init__(
        self, openbis_obj, type=None, data=None, props=None, method=None, **kwargs
    ):
        OpenBisObject.__init__(
            self, openbis_obj, type=type, data=data, props=props, **kwargs
        )
        EntityType.__init__(
            self,
            openbis_obj,
            type=type,
            data=data,
            props=props,
            method=method,
            **kwargs,
        )

    def __dir__(self):
        return [] + EntityType.__dir__(self) + OpenBisObject.__dir__(self)


class MaterialType(
    OpenBisObject,
    EntityType,
    entity="materialType",
    single_item_method_name="get_material_type",
):
    def __init__(
        self, openbis_obj, type=None, data=None, props=None, method=None, **kwargs
    ):
        OpenBisObject.__init__(
            self, openbis_obj, type=type, data=data, props=props, **kwargs
        )
        EntityType.__init__(
            self,
            openbis_obj,
            type=type,
            data=data,
            props=props,
            method=method,
            **kwargs,
        )

    def __dir__(self):
        return [] + EntityType.__dir__(self) + OpenBisObject.__dir__(self)


class ExperimentType(
    OpenBisObject,
    EntityType,
    entity="experimentType",
    single_item_method_name="get_experiment_type",
):
    def __init__(
        self, openbis_obj, type=None, data=None, props=None, method=None, **kwargs
    ):
        OpenBisObject.__init__(
            self, openbis_obj, type=type, data=data, props=props, **kwargs
        )
        EntityType.__init__(
            self,
            openbis_obj,
            type=type,
            data=data,
            props=props,
            method=method,
            **kwargs,
        )

    def __dir__(self):
        return [] + EntityType.__dir__(self) + OpenBisObject.__dir__(self)



from .attribute import AttrHolder
from .definitions import openbis_definitions
from .openbis_object import OpenBisObject
from .property import PropertyHolder
from .utils import VERBOSE


class Sample(OpenBisObject, entity="sample", single_item_method_name="get_sample"):
    """A Sample (new: Object) is one of the most commonly used entities in openBIS."""

    def __init__(
        self, openbis_obj, type, project=None, data=None, props=None, **kwargs
    ):
        self.__dict__["openbis"] = openbis_obj
        self.__dict__["type"] = type
        ph = PropertyHolder(openbis_obj, type)
        self.__dict__["p"] = ph
        self.__dict__["props"] = ph
        self.__dict__["a"] = AttrHolder(openbis_obj, "sample", type)

        if data is not None:
            self._set_data(data)

        # TODO: Why are we using getattr() and setattr() here? They are considerably slower.
        if project is not None:
            self.project = project

        if props is not None:
            for key in props:
                # self.p[key] = props[key]
                setattr(self.p, key, props[key])

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

            if "experiment" in kwargs:
                try:
                    experiment = self.experiment
                    if not "space" in kwargs:
                        project = experiment.project
                        self.a.space = project.space
                except Exception:
                    pass

        if getattr(self, "parents") is None:
            self.a.__dict__["_parents"] = []
        else:
            if not self.is_new:
                self.a.__dict__["_parents_orig"] = self.a.__dict__["_parents"]

        if getattr(self, "children") is None:
            self.a.__dict__["_children"] = []
        else:
            if not self.is_new:
                self.a.__dict__["_children_orig"] = self.a.__dict__["_children"]

        if getattr(self, "components") is None:
            self.a.__dict__["_components"] = []
        else:
            if not self.is_new:
                self.a.__dict__["_components_orig"] = self.a.__dict__["_components"]

    def _set_data(self, data):
        # assign the attribute data to self.a by calling it
        # (invoking the AttrHolder.__call__ function)
        self.a(data)
        self.__dict__["data"] = data

        # put the properties in the self.p namespace (without checking them)
        for key, value in data["properties"].items():
            self.p.__dict__[key.lower()] = value

    def __dir__(self):
        return [
            "type",
            "get_parents()",
            "get_children()",
            "get_components()",
            "add_parents()",
            "add_children()",
            "add_components()",
            "del_parents()",
            "del_children()",
            "del_components()",
            "set_parents()",
            "set_children()",
            "set_components()",
            "get_datasets()",
            "space",
            "project",
            "experiment",
            "container",
            "tags",
            "set_tags()",
            "add_tags()",
            "del_tags()",
            "add_attachment()",
            "get_attachments()",
            "download_attachments()",
            "save()",
            "delete()",
            "mark_to_be_deleted()",
            "unmark_to_be_deleted()",
            "is_marked_to_be_deleted()",
            "attrs",
            "props",
        ] + super().__dir__()

    def _container(self, value=None):
        if value is not None:
            if value == "":
                if self.is_new:
                    pass
                else:
                    self.a.__dict__["_container"] = {}
            else:
                obj = None
                if isinstance(value, str):
                    # fetch object in openBIS, make sure it actually exists
                    obj = getattr(self._openbis, "get_sample")(value)
                elif value is None:
                    self.a.__dict__["_container"] = {}
                else:
                    obj = value

                self.a.__dict__["_container"] = obj.data["identifier"]

                # mark attribute as modified, if it's an existing entity
                if self.is_new:
                    pass
                else:
                    self.a.__dict__["_container"]["isModified"] = True
        else:
            try:
                return self.openbis.get_sample(self.a._container["identifier"])
            except Exception:
                pass

    @property
    def type(self):
        return self.__dict__["type"]

    @type.setter
    def type(self, type_name):
        sample_type = self.openbis.get_sample_type(type_name)
        self.p.__dict__["_type"] = sample_type
        self.a.__dict__["_type"] = sample_type

    def __getattr__(self, name):
        if name in ["container"]:
            return getattr(self, "_" + name)()

        return getattr(self.__dict__["a"], name)

    def __setattr__(self, name, value):
        if name in ["set_properties", "set_tags", "add_tags"]:
            raise ValueError("These are methods which should not be overwritten")

        if name in ["container"]:
            return getattr(self, "_" + name)(value)

        if name in ["p", "props"]:
            if isinstance(value, dict):
                for p in value:
                    setattr(self.__dict__["p"], p, value[p])
            else:
                raise ValueError("please provide a dictionary for setting properties")
        else:
            # must be an attribute in the AttributeHolder class
            setattr(self.__dict__["a"], name, value)

    def _repr_html_(self):
        return self.a._repr_html_()

    def __repr__(self):
        return self.a.__repr__()

    def set_properties(self, properties):
        for prop in properties.keys():
            setattr(self.p, prop, properties[prop])

    set_props = set_properties

    def get_datasets(self, **kwargs):
        return self.openbis.get_datasets(sample=self.permId, **kwargs)

    def get_projects(self, **kwargs):
        return self.openbis.get_project(withSamples=[self.permId], **kwargs)

    @property
    def experiment(self):
        try:
            return self.openbis.get_experiment(self._experiment["identifier"])
        except Exception:
            pass

    def save(self):
        """invoked when code is provided in cases when the type already generates
        the code automatically. In this case, we need to invoke the old V1 method.
        """
        if self.is_new and self.code is not None and self.type.autoGeneratedCode:
            request = self._new_attrs()
            if self.props:
                for prop_name, prop in self.props._property_names.items():
                    if prop["mandatory"]:
                        if (
                            getattr(self.props, prop_name) is None
                            or getattr(self.props, prop_name) == ""
                        ):
                            raise ValueError(
                                "Property '{}' is mandatory and must not be None".format(
                                    prop_name
                                )
                            )

            request = {
                "method": "createReportFromAggregationService",
                "params": [
                    self.openbis.token,
                    self.openbis.get_datastores()["code"][0],
                    "eln-lims-api",
                    {
                        "method": "insertSample",
                        "sampleSpace": self.space.code,
                        "sampleProject": self.project.code,
                        "sampleExperiment": self.experiment.code,
                        "sampleCode": self.code,
                        "sampleType": self.type.code,
                        "sampleProperties": self.props(),
                        "sampleParents": None,
                        "sampleParentsNew": self.parents,
                        "sampleChildrenNew": self.children,
                        "sampleChildrenAdded": [],
                        "sampleChildrenRemoved": [],
                        "changesToDo": [],
                        "sessionToken": self.openbis.token,
                    },
                ],
            }
            resp = self.openbis._post_request(self.openbis.reg_v1, request)
            if VERBOSE:
                print("{} successfully created.".format(self.entity))

            try:
                permid = resp["rows"][0][2]["value"]
            except KeyError:
                print(resp)
            new_entity_data = self.openbis.get_sample(permid, only_data=True)
            self._set_data(new_entity_data)
            return self

        else:
            super().save()



from .property import PropertyHolder
from .attribute import AttrHolder
from .utils import VERBOSE
from .definitions import (
    get_definition_for_entity,
    get_type_for_entity,
    get_method_for_entity,
)
from collections import defaultdict


class OpenBisObject:
    def __init_subclass__(cls, entity=None, single_item_method_name=None):
        """create a specialized parent class.
        The class that inherits from OpenBisObject does not need
        to implement its own __init__ method in order to provide the
        entity name. Instead, it can pass the entity name as a param:
        class XYZ(OpenBisObject, entity="myEntity")
        """
        cls._entity = entity
        cls._single_item_method_name = single_item_method_name

    def __init__(self, openbis_obj, type=None, data=None, props=None, **kwargs):
        self.__dict__["openbis"] = openbis_obj
        self.__dict__["type"] = type
        self.__dict__["p"] = PropertyHolder(openbis_obj, type)
        self.__dict__["a"] = AttrHolder(openbis_obj, self._entity, type)

        # existing OpenBIS object
        if data is not None:
            self._set_data(data)

        if props is not None:
            for key in props:
                setattr(self.p, key, props[key])

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    def __dir__(self):
        defs = get_definition_for_entity(self.entity)
        if self.is_new:
            return defs["attrs_new"]
        else:
            return list(set(defs["attrs"] + defs["attrs_up"]))

    def _set_data(self, data):
        # assign the attribute data to self.a by calling it
        # (invoking the AttrHolder.__call__ function)
        self.a(data)
        self.__dict__["data"] = data

        # put the properties in the self.p namespace (without checking them)
        if "properties" in data:
            for key, value in data["properties"].items():
                self.p.__dict__[key.lower()] = value

        # object is already saved to openBIS, so it is not new anymore
        self.a.__dict__["_is_new"] = False

    @property
    def attrs(self):
        return self.__dict__["a"]

    @property
    def space(self):
        try:
            return self.openbis.get_space(self._space["permId"])
        except Exception:
            pass

    @property
    def project(self):
        try:
            return self.openbis.get_project(self._project["identifier"])
        except Exception:
            pass

    @property
    def experiment(self):
        try:
            return self.openbis.get_experiment(self._experiment["identifier"])
        except Exception:
            pass

    collection = experiment  # Alias

    @property
    def sample(self):
        try:
            return self.openbis.get_sample(self._sample["identifier"])
        except Exception:
            pass

    object = sample  # Alias

    @property
    def _permId(self):
        try:
            return self.data["permId"]
        except Exception:
            return ""

    @property
    def permId(self):
        try:
            return self.data["permId"]["permId"]
        except Exception:
            try:
                return self.a.__dict__["_permId"]["permId"]
            except Exception:
                return ""

    def __getattr__(self, name):
        return getattr(self.__dict__["a"], name)

    def __setattr__(self, name, value):
        if name in ["set_properties", "set_tags", "add_tags"]:
            raise ValueError("These are methods which should not be overwritten")
        setattr(self.__dict__["a"], name, value)

    def _repr_html_(self):
        """Print all the assigned attributes (identifier, tags, etc.) in a nicely formatted table. See
        AttributeHolder class.
        """
        return self.a._repr_html_()

    def __repr__(self):
        """same thing as _repr_html_() but for IPython"""
        return self.a.__repr__()

    def mark_to_be_deleted(self):
        self.__dict__["mark_to_be_deleted"] = True

    def unmark_to_be_deleted(self):
        self.__dict__["mark_to_be_deleted"] = False

    def is_marked_to_be_deleted(self):
        return self.__dict__.get("mark_to_be_deleted", False)

    def delete(self, reason):
        """Delete this openbis entity.
        A reason is mandatory to delete any entity.
        """
        if not self.data:
            return

        self.openbis.delete_openbis_entity(
            entity=self._entity, objectId=self.data["permId"], reason=reason
        )
        if VERBOSE:
            print("{} {} successfully deleted.".format(self._entity, self.permId))

    def _get_single_item_method(self):
        single_item_method = None
        if self._single_item_method_name:
            single_item_method = getattr(self.openbis, self._single_item_method_name)
        else:
            # try to guess the method...
            single_item_method = getattr(self.openbis, "get_" + self.entity)

        return single_item_method

    def save(self):
        get_single_item = self._get_single_item_method()
        # check for mandatory properties before saving the object
        props = None
        if self.props:
            for prop_name, prop in self.props._property_names.items():
                if prop["mandatory"]:
                    if (
                        getattr(self.props, prop_name) is None
                        or getattr(self.props, prop_name) == ""
                    ):
                        raise ValueError(
                            "Property '{}' is mandatory and must not be None".format(
                                prop_name
                            )
                        )

            props = self.p._all_props()

        # NEW
        if self.is_new:
            request = self._new_attrs()
            if props:
                request["params"][1][0]["properties"] = props

            resp = self.openbis._post_request(self.openbis.as_v3, request)

            if VERBOSE:
                print("{} successfully created.".format(self.entity))
            new_entity_data = get_single_item(resp[0]["permId"], only_data=True)
            self._set_data(new_entity_data)
            return self

        # UPDATE
        else:
            request = self._up_attrs(method_name=None, permId=self._permId)

            if props:
                request["params"][1][0]["properties"] = props

            resp = self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("{} successfully updated.".format(self.entity))
            new_entity_data = get_single_item(
                self.permId, only_data=True, use_cache=False
            )
            self._set_data(new_entity_data)
            return self


class Transaction:
    def __init__(self, *entities):

        self.entities = {}
        self.reason = "no reason"

        if not entities:
            return

        for entity in entities:
            self.add(entity)

    def add(self, entity_obj):
        """Collect the creations or updates of entities.
        self.entities = {
            "sample": {
                {
                    "create": [...],
                    "update": [...]
                },
            },
            "dataSet": {
                {
                    "update": [...]
                }
            }
        }
        """

        if not entity_obj.entity in self.entities:
            self.entities[entity_obj.entity] = defaultdict(list)

        mode = "update"
        if entity_obj.is_new:
            mode = "create"
        elif entity_obj.is_marked_to_be_deleted():
            mode = "delete"
        else:
            mode = "update"

        self.entities[entity_obj.entity][mode].append(entity_obj)

    def commit(self):
        """Merge the individual requests to Make one single request.
        For each entity-type and mode (create, update) an individual request will be sent,
        as the method name differs.
        """

        import copy

        for entity_type in self.entities:
            for mode in self.entities[entity_type]:

                request_coll = []
                for entity in self.entities[entity_type][mode]:
                    if mode == "delete":
                        delete_options = get_type_for_entity(entity_type, "delete")
                        delete_options["reason"] = self.reason
                        method = get_method_for_entity(entity_type, "delete")
                        request = {
                            "method": method,
                            "params": [
                                entity.openbis.token,
                                [entity.data["permId"]],
                                delete_options,
                            ],
                        }
                        request_coll.append(request)
                        continue
                    props = None
                    if entity.props:
                        for prop_name, prop in entity.props._property_names.items():
                            if prop["mandatory"]:
                                if (
                                    getattr(entity.props, prop_name) is None
                                    or getattr(entity.props, prop_name) == ""
                                ):
                                    raise ValueError(
                                        "Property '{}' is mandatory and must not be None".format(
                                            prop_name
                                        )
                                    )
                    props = entity.p._all_props()

                    if mode == "create":
                        request = entity._new_attrs()
                        if props:
                            request["params"][1][0]["properties"] = props

                    elif mode == "update":
                        request = entity._up_attrs(
                            method_name=None, permId=entity._permId
                        )
                        if props:
                            request["params"][1][0]["properties"] = props

                    else:
                        raise ValueError(f"Unkown mode: {mode}")

                    request_coll.append(request)

                if request_coll:
                    # copy the first item of all requests
                    batch_request = copy.deepcopy(request_coll[0])
                    # merge all requests into one
                    for i, request in enumerate(request_coll):
                        if i == 0:
                            continue
                        batch_request["params"][1].append(request["params"][1][0])

                    try:
                        resp = entity.openbis._post_request(
                            entity.openbis.as_v3, batch_request
                        )
                        if VERBOSE:
                            print(f"{i+1} {entity_type}(s) {mode}d.")

                        # mark every sample as not being new anymore
                        # and add the permId attribute received by the response
                        # we assume the response permIds are the same order as we sent them
                        if resp:
                            for i, resp_item in enumerate(resp):
                                if mode == "delete":
                                    continue
                                self.entities[entity_type][mode][i].a.__dict__[
                                    "_is_new"
                                ] = False
                                self.entities[entity_type][mode][i].a.__dict__[
                                    "_permId"
                                ] = resp_item

                    except ValueError as err:
                        if VERBOSE:
                            print(f"ERROR: {mode} of {i+1} {entity_type}(s) FAILED")
                        raise ValueError(err)



from .attribute import AttrHolder
from .openbis_object import OpenBisObject 
from .utils import VERBOSE

class RoleAssignment(OpenBisObject):
    """ managing openBIS role assignments
    """

    def __init__(self, openbis_obj, data=None, **kwargs):
        self.__dict__['openbis'] = openbis_obj
        self.__dict__['a'] = AttrHolder(openbis_obj, 'roleAssignment' )

        if data is not None:
            self.a(data)
            self.__dict__['data'] = data

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])


    def __dir__(self):
        """all the available methods and attributes that should be displayed
        when using the autocompletion feature (TAB) in Jupyter
        """
        return [
            'id', 'role', 'roleLevel', 'space', 'project','group'
        ]

    def __str__(self):
        return "{}".format(self.get('role'))

    def delete(self, reason='no reason specified'):
        self.openbis.delete_openbis_entity(
            entity='roleAssignment',
            objectId=self._id,
            reason= reason
        )
        if VERBOSE:
            print("RoleAssignment role={}, roleLevel={} successfully deleted.".format(self.role, self.roleLevel))



import site
import sys

site.ENABLE_USER_SITE = "--user" in sys.argv[1:]
if sys.version_info < (3, 6):
    sys.exit("Sorry, Python < 3.6 is not supported")

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="PyBIS",
    version="1.31.2",
    author="Swen Vermeul • ID SIS • ETH Zürich",
    author_email="swen@ethz.ch",
    description="openBIS connection and interaction, optimized for using with Jupyter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://sissource.ethz.ch/sispub/openbis/tree/master/pybis",
    packages=find_packages(),
    license="Apache Software License Version 2.0",
    install_requires=[
        "pytest",
        "requests",
        "urllib3",
        "pandas",
        "click",
        "texttable",
        "tabulate",
    ],
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
